# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Distributed tuning with pluggable backends.

:func:`pyglove.iter` provides an interface for sampling examples from a search
space within a process. To support distributed tuning, PyGlove introduces
:func:`pyglove.sample`, which is almost identical but with more features:

 * Allow multiple worker processes (aka. workers) to collaborate on a search
   with failover handling.
 * Each worker can process different trials, or can cowork on the same trials
   via work groups.
 * Provide APIs for communicating between the co-workers.
 * Provide API for retrieving the search results.
 * Provide a pluggable backend system for supporting user infrastructures.

"""

# pylint: disable=g-bad-import-order

# User facing APIs for tuning.
from pyglove.core.tuning.sample import sample
from pyglove.core.tuning.backend import poll_result

from pyglove.core.tuning.backend import default_backend
from pyglove.core.tuning.backend import set_default_backend

# Tuning protocols.
from pyglove.core.tuning.protocols import Measurement
from pyglove.core.tuning.protocols import Trial
from pyglove.core.tuning.protocols import Result
from pyglove.core.tuning.protocols import Feedback
from pyglove.core.tuning.protocols import RaceConditionError

# Interface for early stopping.
from pyglove.core.tuning.early_stopping import EarlyStoppingPolicy

# Interfaces for tuning backend developers.
from pyglove.core.tuning.backend import Backend
from pyglove.core.tuning.backend import add_backend
from pyglove.core.tuning.backend import available_backends

# Importing local backend.
import pyglove.core.tuning.local_backend

# pylint: enable=g-bad-import-order



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain algo copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.geno.DNA."""

import unittest

from pyglove.core.geno.base import DNA
from pyglove.core.geno.categorical import oneof
from pyglove.core.geno.dna_generator import dna_generator
from pyglove.core.geno.dna_generator import DNAGenerator
from pyglove.core.geno.space import constant


class DNAGeneratorTest(unittest.TestCase):
  """Tests for DNAGenerator base."""

  def _dna_spec(self):
    return oneof([constant(), constant()])

  def test_generator_without_feedback(self):

    # Test for DNAGenerator that does not need feedbacks.
    class DummyGenerator(DNAGenerator):

      def _propose(self):
        return DNA(0)

    algo = DummyGenerator()
    algo.setup(self._dna_spec)
    self.assertFalse(algo.needs_feedback)
    self.assertFalse(algo.multi_objective)
    self.assertEqual(algo.num_proposals, 0)
    dna = algo.propose()
    self.assertEqual(dna, DNA(0))
    self.assertEqual(algo.num_proposals, 1)
    self.assertEqual(algo.num_feedbacks, 0)
    algo.feedback(dna, 0)
    self.assertEqual(algo.num_feedbacks, 1)

  def test_single_objective_optimizer(self):

    class DummySingleObjectiveOptimizer(DNAGenerator):

      def _setup(self):
        self.max_reward = None

      def _propose(self):
        return DNA(1)

      def _feedback(self, dna, reward):
        if self.max_reward is None or reward > self.max_reward:
          self.max_reward = reward

    algo = DummySingleObjectiveOptimizer()
    algo.setup(self._dna_spec)
    self.assertTrue(algo.needs_feedback)
    self.assertFalse(algo.multi_objective)
    self.assertEqual(algo.num_proposals, 0)
    dna = algo.propose()
    self.assertEqual(dna, DNA(1))
    self.assertEqual(algo.num_proposals, 1)
    self.assertEqual(algo.num_feedbacks, 0)
    algo.feedback(dna, 1.2)
    self.assertEqual(algo.num_feedbacks, 1)
    self.assertEqual(algo.max_reward, 1.2)

    with self.assertRaisesRegex(
        ValueError,
        '.* is single objective, but the reward .*'
        'contains multiple objectives'):
      algo.feedback(dna, (0, 1))

  def test_multi_objective_optimizer(self):

    class DummyMultiObjectiveOptimizer(DNAGenerator):

      @property
      def multi_objective(self):
        return True

      def _setup(self):
        self.rewards = []

      def _propose(self):
        return DNA(1)

      def _feedback(self, dna, reward):
        self.rewards.append(reward)

    algo = DummyMultiObjectiveOptimizer()
    algo.setup(self._dna_spec)
    self.assertTrue(algo.needs_feedback)
    self.assertTrue(algo.multi_objective)
    self.assertEqual(algo.num_proposals, 0)
    dna = algo.propose()
    self.assertEqual(dna, DNA(1))
    self.assertEqual(algo.num_proposals, 1)
    self.assertEqual(algo.num_feedbacks, 0)
    algo.feedback(dna, (0.9, 1.2))
    self.assertEqual(algo.num_feedbacks, 1)
    self.assertEqual(algo.rewards, [(0.9, 1.2)])
    algo.feedback(dna, 1.)
    self.assertEqual(algo.num_feedbacks, 2)
    self.assertEqual(algo.rewards, [(0.9, 1.2), (1.,)])


class DNAGeneratorDecoratorTest(unittest.TestCase):

  def test_dna_generator_decorator(self):

    @dna_generator
    def first_ten(dna_spec):
      dna = None
      for _ in range(10):
        dna = dna_spec.next_dna(dna)
        if dna is None:
          break
        yield dna

    algo = first_ten()    # pylint: disable=no-value-for-parameter
    dna_spec = oneof([constant()] * 8)
    algo.setup(dna_spec)

    dna_list = []
    with self.assertRaises(StopIteration):
      while True:
        dna_list.append(algo.propose())
    self.assertEqual(
        dna_list,
        [DNA(i) for i in range(8)])

    # Test error preservation from early proposal.
    @dna_generator
    def bad_generator(unused_spec):
      if True:  # pylint: disable=using-constant-test
        raise ValueError('bad initializer')
      yield DNA(0)

    algo = bad_generator.partial()
    algo.setup(None)

    with self.assertRaisesRegex(
        ValueError, 'bad initializer'):
      algo.propose()

    with self.assertRaisesRegex(
        ValueError, 'Error happened earlier: bad initializer'):
      algo.propose()


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sweeping DNA generator."""

from typing import Any

from pyglove.core.geno.base import DNA
from pyglove.core.geno.dna_generator import DNAGenerator


class Sweeping(DNAGenerator):
  """Sweeping (Grid Search) DNA generator."""

  def _setup(self):
    """Setup DNA spec."""
    self._last_proposed_dna = None

  def _propose(self) -> DNA:
    """Propose a random DNA."""
    next_dna = self.dna_spec.next_dna(self._last_proposed_dna)
    if next_dna is None:
      raise StopIteration()
    self._last_proposed_dna = next_dna
    return next_dna

  def _replay(self, trial_id: int, dna: DNA, reward: Any) -> None:
    """Replay the history to recover the last proposed DNA."""
    del trial_id, reward
    self._last_proposed_dna = dna
